#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <strfunc.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sysexits.h>
#include <ctype.h>
#include <mbox.h>
#include <mail.h>
#include <lgba.h>
#include "common.h"

#ifndef CONFIG_DIR
#define CONFIG_DIR	"./"
#endif

info *Info = NULL;

void
PlainError(const char *msg) {
	printf("Content-type: text/plain\n\n%s\n", msg);
	exit(2);
}

static void
_on_memory_failure() {
	static char msg[] = "Memory allocation failed\n";
	(void)write(2, msg, sizeof(msg) - 1);
	(void)write(1, msg, sizeof(msg) - 1);
	exit(EX_OSERR);
}

void SetupSorting(info *Info, int *saveit);

void
init() {
	char *p;

	/*
	 * This program intended to run under HTTP daemon.
	 */
	if(getuid() > 400) {
		printf("Trying to hack me, bad motherlover?\n");
		exit(2);
	}

	if(strfunc_version() < 746) {
		printf("Content-type: text/html\n\n"
			"Please upgrade libstrfunc version to "
			"at least 7.4.5!<BR>\n"
			"It can be found at "
			"<A HREF=http://www.spelio.net.ru/soft"
			">www.spelio.net.ru/soft</A>\n"
		);
		exit(2);
	}

	/*
	 * Print something and exit.
	 */
	strfunc_memory_failure_callback(_on_memory_failure);

	Info = (info *)sf_calloc(1, sizeof(info));

	Info->username = getenv("REMOTE_USER");
	if(Info->username == NULL
		|| Info->username[0] == '\0'
		|| Info->username[0] == '-'
	) {
		PlainError("Authentication error!");
	}

	for(p = Info->username; *p; p++) {
		*p = tolower(*p);
		if(*p == '/' || *p == ' ') {
			/* Very strange user name, huh? */
			PlainError("Authentication error!");
		}
	}

   	Info->pgp_passphrase = param("pgppassphrase");

#ifndef NO_LGBA

	if(lgba_init(NULL) || lgba_open())
		PlainError("Can't init LGBA");
    
	/* Set necessary settings */
	lgba_setuser("root");

	/*
	 * Attributes we're interested in.
	 */
	char *avl[]={
		"cn",
		"exaview-sortby",
		"exaview-lang",
		"pgp-pubring",
		"pgp-secring",
		"email-signature",
		NULL};

	lgba_setattrlist(avl);

	if(load(Info->username)) {
		printf("Content-type: text/plain\n\n");
		printf("LGBA Authorization error!\n");
	}

	int saveit = 0;

	SetupSorting(Info, &saveit);

	// Setup language; read the configuration file
	char cfile[sizeof(CONFIG_DIR) + 64];
	sprintf(cfile, CONFIG_DIR "/exaview.conf");

	slist *langprefs = getlanguageprefs();
	char *lang;
	if((lang=param("lang")) && strlen(lang) < 10) {
		setattr("exaview-lang", lang);
		saveit=1;
	} else {
		lang=getattr("exaview-lang");
	}

	if(*lang && (strspn(lang, "abcdefghijklmnopqrstuvwxyABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789") == strlen(lang))) {
		if(!langprefs)
			langprefs = sinit();
		// Insert
		sins(langprefs, lang, 0);
	}

	if(langprefs) {
		char *cfe=cfile + strlen(cfile);
		size_t i;
		for(i=0; i < langprefs->count; i++) {
			// Security check
			if(langprefs->lens[i] > 10)
				continue;
			sprintf(cfe, ".%s", langprefs->list[i]);
			if(cfgread(cfile) == 0)
				goto rcf;   // Success
		}
		*cfe=0;
		if(i == langprefs->count) {
			if(cfgread(cfile))
				PlainError("Can't parse config/form");
		}
	} else {
		if(cfgread(cfile))
			PlainError("Can't parse config/form");
	}

rcf:

	if(saveit)
		lgba_save(0);

	char *domain = cfgget("domain");
	// Constructing the "From" field.
	p = (char *)sf_malloc(
		strlen(lgba_ObjectAttribute()) * 2
		+ (domain?(strlen(domain)):0)
		+ 64
	);

	sprintf(p, "${cn?${cn[0]}:${~[0]}} <${~[0]}%s%s>",
			domain?"@":"",
			domain?domain:""
	);
	Info->fullname = sf_strdup(getformat(p));

	sprintf(p, "${~[0]}%s%s",
			domain?"@":"",
			domain?domain:""
	);
	Info->from = sf_strdup(getformat(p));

	free(p);

#else

	Info->sortby=NULL;
	Info->from=Info->username;

	if(cfgread(CONFIG_DIR "/exaview.conf"))
		PlainError("Can't parse config/form");

#endif

	Info->msgnum = -1;
	Info->msgid = -1;
	Info->partid = -1;

	int msgid_provided = 0;

	if((p=param("msgid")))  { Info->msgid=atoi(p); msgid_provided = 1; }
	if((p=param("msgnum"))) Info->msgnum=atoi(p);
	if((p=param("partid"))) Info->partid=atoi(p);

	Info->script_name = getenv("SCRIPT_NAME");
	if(Info->script_name) {
		p = strrchr(Info->script_name, '/');
		if(p) Info->script_name = p + 1;
		Info->script_name = strdup(Info->script_name);
	}

	Info->path_info = getenv("PATH_INFO");
	if(Info->path_info) {
		if(!msgid_provided) {
			if(strchr(Info->path_info + 1, '/')) {
				Info->msgid = atoi(Info->path_info + 1);
			}
		}
		p = strrchr(Info->path_info, '/');
		if(p) Info->path_info = p + 1;
		Info->path_info = sf_strdup(url_decode(Info->path_info));

		if(!msgid_provided && Info->msgid != -1) {
			if(Info->script_name
			&& strcmp(Info->script_name, Info->path_info) == 0)
				Info->msgid = -1;
		}
	}

}



void
SetupSorting(info *Info, int *saveit) {
	char *p;

	// Setup sorting.
	Info->sortby = getattr("exaview-sortby");

	// If something was passed as the CGI param:
	p = param("sortby");
	if(p == NULL)
		return;

	// If we would like to reverse the order.
	if(ecq(Info->sortby, p)
		|| (*Info->sortby == '-' && ecq(Info->sortby+1, p))) {

		if(*Info->sortby == '-') {
			Info->sortby++;
		} else {
			Info->sortby=(char *)sf_malloc(strlen(p) + 2);
			*Info->sortby='-';
			strcpy(Info->sortby + 1, p);
		}
       	 } else {
		if(*Info->sortby == '-') {
			Info->sortby = (char *)sf_malloc(strlen(p) + 2);
			*Info->sortby='-';
			strcpy(Info->sortby+1, p);
		} else {
			Info->sortby=p;
		}
	}

	setattr("exaview-sortby", Info->sortby);
	*saveit = 1;
}
