#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <strfunc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>
#include <lgba.h>

#include <common.h>
#include <mbox.h>

#include "pgpBase.h"
#include "pgpEncode.h"
#include "pgpPFLErrors.h"
#include "pgpUtilities.h"
#include "pgpFileSpec.h"
#include "pgpUtilitiesPriv.h"
#include "pgpRandomPool.h"
#include "pgpSDKPrefs.h"
#include "pgpKeys.h"
#include "pgpErrors.h"

int  pgpinit(char *recipient);
char *pgpencode(void *data, size_t size, int op=1);
void pgpcleanup();

inline const char *cf(char *a) {
	register char *z=cfgget(a);
	return z?z:"";
};

PGPError		err = kPGPError_NoErr;
PGPContextRef	context = NULL;
PGPKeySetRef	keyringset = NULL;
PGPKeySetRef	resultset = NULL;
PGPFilterRef	filter = NULL;
PGPFileSpecRef pubFileSpec;
PGPFileSpecRef secFileSpec;
PGPKeyListRef foundKeysList = kInvalidPGPKeyListRef;
PGPKeyIterRef keyListIterator= kInvalidPGPKeyIterRef;
PGPKeyRef	signingKey = kInvalidPGPKeyRef;
static char buf[256];
void *outbuf=NULL;
PGPSize outbuflen=0;
int pgp_RELOAD=0;

char *pgpErrString=NULL;

PGPError pgpMyEventHandler(PGPContextRef context, struct PGPEvent *event, PGPUserValue userValue);

int pgpinit(char *recipient) {
	PGPUInt32 numKeys=0;

	pgp_RELOAD=0;

	/* Initialize the PGPsdk libs */
	err = PGPsdkInit();
	if(IsPGPError(err)) return -1;

	/* Create the PGPContext */
	err = PGPNewContext( kPGPsdkAPIVersion, &context );
	if(IsPGPError(err)) return -1;

	struct lenbuf *lbp, *lbs;
	int fd;

	lbp=getbinval("pgp-pubring", 0);
	lbs=getbinval("pgp-secring", 0);
	if(!lbp || !lbs || !lbp->buf || !lbs->buf) {
		pgpErrString=sf_strdup(cf("pgp-not-configured"));
		return -1;
	};

	sprintf(buf, "/tmp/exaview-%s-%d-pr-XXXXXX",
		getformat("${object[0]}"), getpid());
	if((fd=mkstemp(buf)) == -1) return -3;
	write(fd, lbp->buf, lbp->len);
	close(fd);

	err = PGPNewFileSpecFromFullPath( context, buf, &pubFileSpec );
	if(IsPGPError(err)) { unlink(buf); return -1; };

	sprintf(buf, "/tmp/exaview-%s-%d-sr-XXXXXX",
		getformat("${object[0]}"), getpid());
	if((fd=mkstemp(buf)) == -1) return -3;
	
	if(lbs && lbs->len > 0) write(fd, lbs->buf, lbs->len);
	close(fd);

	err = PGPNewFileSpecFromFullPath( context, buf, &secFileSpec );
	if(IsPGPError(err)) {
		char *fp;
		unlink(buf);
		PGPGetFullPathFromFileSpec(pubFileSpec, &fp);
		if(!IsPGPError(err)) {
			unlink(fp);
			sprintf(buf, "%s-bak-1", fp);
			unlink(buf);
		};
		return -1;
	};

	err = PGPOpenKeyRingPair ( context, kPGPKeyRingOpenFlags_Mutable,
		secFileSpec, pubFileSpec, &keyringset );

	/* Signing key */
	snprintf(buf, sizeof(buf), "<${object[0]}@%s%s",
		cf("domain"), *cf("domain")?">":"");
	buf[sizeof(buf)-1]=0;
	err = PGPNewUserIDStringFilter( context, getformat(buf),
		kPGPMatchSubString, &filter );

	if( IsPGPError(err) ) return -1;
	err = PGPFilterKeySet( keyringset, filter, &resultset );
	if( IsPGPError(err) ) return -1;

	err = PGPCountKeys( resultset, &numKeys );
	if( IsPGPError(err) ) return -1;
	if(numKeys > 0) {
/*
		err = PGPGetDefaultPrivateKey( keyringset, &signingKey );
		if( IsPGPError(err) ) {
			PGPGetErrorString( err, sizeof(buf), buf );
			pgpErrString=sf_strdup(buf);
			return -1;
		};
*/

		err = PGPOrderKeySet(keyringset, kPGPAnyOrdering,
			&foundKeysList);
		if( IsPGPError(err) ) return -1;
		err = PGPNewKeyIter( foundKeysList, &keyListIterator );
		if( IsPGPError(err) ) return -1;
		err = PGPKeyIterNext( keyListIterator, &signingKey );
		if( IsPGPError(err) ) return -1;

		PGPBoolean isSecret=0;

		while ( signingKey ) {
			err = PGPGetKeyBoolean( signingKey, kPGPKeyPropIsSecret,
				&isSecret);               
			if(IsPGPError(err)) {
				err = PGPKeyIterNext(keyListIterator,
					&signingKey );
				if(IsPGPError(err)) break;
				continue;
			};
			if(isSecret) break;

			err = PGPKeyIterNext( keyListIterator, &signingKey );
			if(IsPGPError(err)) break;
		};
		if(!isSecret) {
			pgpErrString=sf_strdup("Can't get default private key.\n");
			return -1;
		};

	};

	PGPFreeFilter( filter );
	PGPFreeKeySet( resultset );

	/* other keys */
	if(recipient[0] == '0' && recipient[1] == 'x') {
		PGPKeyID keyid;
		PGPGetKeyIDFromString(recipient, &keyid);
		err = PGPNewKeyIDFilter( context, &keyid, &filter );
	} else {
		err = PGPNewUserIDStringFilter( context, recipient,
		kPGPMatchSubString, &filter );
	};

	if(IsPGPError(err)) return -1;
	err = PGPFilterKeySet( keyringset, filter, &resultset );
	if(IsPGPError(err)) return -1;

	PGPFreeFilter ( filter );

	srandomdev();
	int i;
	while(!PGPGlobalRandomPoolHasMinimumEntropy())
		for(i=0; i<50; i++)
			PGPGlobalRandomPoolAddKeystroke(random());

	return 0;
};

char *pgpencode(void *data, size_t size, int op = 1) {
	char mimeSeparator [ kPGPMimeSeparatorSize ];
	PGPSize mimeBodyOffset;

	// Encrypt flag
	if(op & 1) {
		PGPUInt32 numKeys=0;
		err = PGPCountKeys( resultset, &numKeys );
		if(numKeys < 1) {
			if(pgpErrString) free(pgpErrString);
			pgpErrString = sf_strdup(cf("pgp-no-public-key"));
			return NULL;
		};
	};

	// Sign flag
	if(op & 2) {
		if(signingKey == kInvalidPGPKeyRef) {
			if(pgpErrString) free(pgpErrString);
			pgpErrString = sf_strdup(cf("pgp-no-secret-key"));
			return NULL;
		};
	};

	err = PGPEncode( context,
		PGPOArmorOutput( context, TRUE ),
		PGPOInputBuffer( context, data, size),
		PGPOCompression( context, TRUE ),
		PGPOAllocatedOutputBuffer(context, &outbuf, 10*1024*1024, &outbuflen),

		PGPOEventHandler( context, pgpMyEventHandler, NULL),

		(op & 0x04)?
			PGPOPGPMIMEEncoding( context, TRUE, &mimeBodyOffset, mimeSeparator)
			:PGPONullOption( context ),

		(op & 0x08)?
			PGPODetachedSig( context, PGPOLastOption( context ) )
			:PGPONullOption( context ),

		(op & 0x04)?
			PGPOOmitMIMEVersion( context, TRUE)
			:PGPONullOption( context ),

		(op & 0x01)?
			PGPOEncryptToKeySet( context, resultset )
			:PGPONullOption( context ),

		(op & 0x02)?
			PGPOSignWithKey(context, signingKey,
				PGPOPassphrase( context, Info->pgp_passphrase ),
				PGPOLastOption( context ) ):
			PGPONullOption(context),

		(op & 0x02)?
			PGPOPassphrase( context, Info->pgp_passphrase ):
			PGPONullOption(context),

		PGPOVersionString(context, "Exaview version 4.0/pgp (c) 1998-2000 <vlm@spelio.net.ru>"),

		PGPOLastOption( context ) );

	if(IsPGPError(err)) {
		PGPGetErrorString( err, sizeof(buf), buf );
		pgpErrString=sf_strdup(cf(buf));
		return NULL;
	};

	return (char *)outbuf;
};

void pgpcleanup() {
	char *fp;

	if( PGPFileSpecRefIsValid( pubFileSpec ) ) {
		PGPFreeFileSpec( pubFileSpec );
		err = PGPGetFullPathFromFileSpec(pubFileSpec, &fp);
		if(!IsPGPError(err)) {
			if(pgp_RELOAD)
				setattr("pgp-pubring", fp);
			unlink(fp);
			sprintf(buf, "%s-bak-1", fp);
			unlink(buf);
			sprintf(buf, "%s-bak-2", fp);
			unlink(buf);
		};
	};

	if( PGPFileSpecRefIsValid( secFileSpec ) ) {
		PGPFreeFileSpec( secFileSpec );
		err = PGPGetFullPathFromFileSpec(secFileSpec, &fp);
		if(!IsPGPError(err)) {
			if(pgp_RELOAD) setattr("pgp-secring", fp);
			unlink(fp);
			sprintf(buf, "%s-bak-1", fp);
			unlink(buf);
			sprintf(buf, "%s-bak-2", fp);
			unlink(buf);
		};
	};

	if(pgp_RELOAD) {
		lgba_saveloaded(0);
		pgp_RELOAD=0;
	};

	if( PGPKeySetRefIsValid( resultset ) )
		PGPFreeKeySet( resultset );
	if( PGPKeySetRefIsValid( keyringset ) )
		PGPFreeKeySet( keyringset );
	if( PGPContextRefIsValid( context ) )
		PGPFreeContext( context );
	PGPsdkCleanup();
};

void *pgpdecode(void *data, size_t *size,
	char *optDetachedSignature = NULL,
	int *optSigningResult = NULL
)
{
	void *outbuf=NULL;
	PGPSize outbuflen=0;
	PGPUInt32 numKeys=0;
	size_t tmpsize;

	if(!data) return NULL;

	if(!size) {
		tmpsize = strlen((char *)data);
		size = &tmpsize;
	};

	err = PGPCountKeys( keyringset, &numKeys );
	if(numKeys < 1) {
		if(pgpErrString)
			free(pgpErrString);
		pgpErrString = sf_strdup(cf("pgp-no-secret-key"));
		return NULL;
	};

	err = PGPDecode( context,
		PGPOKeySetRef( context, keyringset ),
		PGPOEventHandler( context, pgpMyEventHandler, optSigningResult),
		PGPOInputBuffer( context, data, *size),
		PGPOAllocatedOutputBuffer( context, &outbuf,
			10 * 1024 * 1024, &outbuflen ),

		Info->pgp_passphrase?
			PGPOPassphrase( context, Info->pgp_passphrase )
			:PGPONullOption( context ),
		
/*
		optDetachedSignature?
			PGPORawPGPInput ( context, TRUE )
			:PGPONullOption( context ),
*/

		optDetachedSignature?
			PGPODetachedSig( context,
				PGPOInputBuffer( context,
					optDetachedSignature, strlen(optDetachedSignature)),
				PGPOLastOption( context ) )
			:PGPONullOption( context ),

		PGPOLastOption( context )
	);

	if(IsPGPError(err)) {
//		PGPGetErrorString( err, sizeof(buf), buf );
//		pgpErrString=sf_strdup(buf);
		return NULL;
	};

	*size=outbuflen;
	return outbuf;
};

char **
pgp_c2userinfo(char *field, void *) {
	return getattr2(field);
};

char **
pgp_c2pgpinfo(char *field, void *keyPassed) {
	register PGPKeyRef key=(PGPKeyRef)keyPassed;

	static char buf[kPGPMaxKeyIDStringSize>512?kPGPMaxKeyIDStringSize:512];
	static char *c2[]={NULL, NULL};
	PGPSize size;
	PGPTime ptime;
	int keynumber;
	PGPUInt16 yyyy, mm, dd;
	PGPBoolean pbool;


	if(*field == 'i' || *field == 'I') {
		PGPKeyID keyid;
		err = PGPGetKeyIDFromKey(key, &keyid);
		if(IsPGPError(err)) return NULL;
		PGPGetKeyIDString(&keyid,
		(*field=='I')?kPGPKeyIDString_Full:kPGPKeyIDString_Abbreviated, buf);
		*c2=buf;
		return c2;
	};


	if(*field == 'n' || (*field == 'e' && field[1] != 'x') || *field == 'u') {
		PGPUserIDPropName pn;
		if(*field == 'n') pn=kPGPUserIDPropCommonName;
		else if(*field == 'u') pn=kPGPUserIDPropName;
		else pn=kPGPUserIDPropEmailAddress;

		PGPUserIDRef uidRef;
		PGPGetPrimaryUserID( key, &uidRef );
		err = PGPGetUserIDStringBuffer( uidRef, pn, sizeof(buf), buf, &size);
		if(IsPGPError(err)) {
			strcpy(buf, "[error: ");
			PGPGetErrorString(err, sizeof(buf-10), buf+8);
			strcat(buf, "]");
			*c2=buf;
			return c2;
		};
		buf[size]=0;
		*c2=html_quote(buf);
		return c2;
	};

	if(*field == 't') {
		PGPBoolean isSecret;
		err = PGPGetKeyBoolean( key, kPGPKeyPropIsSecret, &isSecret);
		if(IsPGPError(err)) { *c2="[error]"; return c2; };
		if(isSecret) *c2="Secret";
		else *c2="Public";
		return c2;
	};

	if(*field == 'b') {
   		err = PGPGetKeyNumber( key, kPGPKeyPropBits, &keynumber );
		if(IsPGPError(err)) { *c2="[error]"; return c2; };
		sprintf(buf, "%u", keynumber);
		*c2=buf;
		return c2;
	};

	if(*field == 'c') {
		err = PGPGetKeyTime ( key, kPGPKeyPropCreation, &ptime );
		if(IsPGPError(err)) { *c2="[error]"; return c2; };
   		PGPGetYMDFromPGPTime( ptime, &yyyy, &mm, &dd);
		sprintf(buf, "%02d/%02d/%04d", dd, mm, yyyy);
		*c2=buf;
		return c2;
	};

	if(*field == 'a') {
		err = PGPGetKeyNumber(key, kPGPKeyPropAlgID, &keynumber );
		switch( keynumber ) {
		case kPGPPublicKeyAlgorithm_RSA:
			*c2="RSA"; break;
		case kPGPPublicKeyAlgorithm_RSAEncryptOnly:
			*c2="RSA encrypt"; break;
		case kPGPPublicKeyAlgorithm_RSASignOnly:
			*c2="RSA sign"; break;
		case kPGPPublicKeyAlgorithm_ElGamal:
			*c2="DH"; break;
		case kPGPPublicKeyAlgorithm_DSA:
			*c2=*c2="DSS"; break;
		default:
			*c2="""Unknown";
		};
		return c2;
	};

	if(*field == 'o') {
		*buf=0;
		PGPBoolean isrev,isdis,cansign,canencrypt;
		PGPGetKeyBoolean( key, kPGPKeyPropIsRevoked, &isrev );
		if(isrev)
			strcat(buf, "Revoked\n");
		PGPGetKeyBoolean( key, kPGPKeyPropIsDisabled, &isdis );
		if(isdis)
			strcat(buf, "Disabled\n");
		PGPGetKeyBoolean( key, kPGPKeyPropCanEncrypt, &canencrypt );
		PGPGetKeyBoolean( key, kPGPKeyPropCanSign, &cansign );
		if(cansign && canencrypt) {
			strcat(buf, "Key can sign and may be used for encryption\n");
		} else {
			if(cansign)
				strcat(buf, "Key can sign\n");
			if(canencrypt)
				strcat(buf, "Key may be used for encryption\n");
			else strcat(buf, "Users cannot encrypt to this key\n");
		};

		*c2=buf;
		return c2;
	};

	if(!strncasecmp(field, "exp", 3)) {
        PGPGetKeyBoolean( key, kPGPKeyPropIsExpired, &pbool );
        if(pbool) { *c2="Expired"; return c2; };
		err = PGPGetKeyTime( key, kPGPKeyPropExpiration, &ptime );
		if(IsPGPError(err)) { *c2="[error]"; return c2; };
		if(ptime != kPGPExpirationTime_Never) {
			PGPGetYMDFromPGPTime( ptime , &yyyy, &mm, &dd);
			sprintf(buf, "%02d/%02d/%04d", dd, mm, yyyy);
		} else *c2="Never";
		return c2;
	};


	return NULL;
};

void pgpprintkeys(char *fmt) {
	PGPKeyRef key = kInvalidPGPKeyRef;

	PGPOrderKeySet( keyringset, kPGPAnyOrdering, &foundKeysList );
	if( IsPGPError(err) ) return;
	err = PGPNewKeyIter( foundKeysList, &keyListIterator );

	static fmt_base *fb=NULL;

	if(!fb) {
		fb=format_init();
		format_metarule(fb, '{', '}', pgp_c2userinfo);
		format_metarule(fb, '`', '\'', pgp_c2pgpinfo);
	};

	while(1) {
		err = PGPKeyIterNext( keyListIterator, &key );
		if(IsPGPError(err)) break;
		printf("%s", formatf(fb, fmt, key));
	};

};

int pgpkeyinit(int keyalg, int keysize, int expirationDays ) {
	PGPKeyRef masterkey = kInvalidPGPKeyRef;
	PGPSubKeyRef subkey = kInvalidPGPSubKeyRef;
	PGPPublicKeyAlgorithm sigalg, encalg;
	struct lenbuf *lbp, *lbs;
	int fd;

	if(pgpErrString) { free(pgpErrString); pgpErrString=NULL; };

	lbp=getbinval("pgp-pubring", 0);
	lbs=getbinval("pgp-secring", 0);

	if((lbp && lbp->buf) || (lbs && lbs->buf)) {
		pgpErrString=sf_strdup("Keys alredy initialized.");
		return -1;
	};

	if(keysize < 768) keysize = 768;

	switch(keyalg) {
		case 0:
    			sigalg = kPGPPublicKeyAlgorithm_DSA;
    			encalg = kPGPPublicKeyAlgorithm_ElGamal;
			if(keysize > 4096) keysize = 4096;
			break;
		case 1:
			sigalg=kPGPPublicKeyAlgorithm_RSA;
			encalg=kPGPPublicKeyAlgorithm_RSA;
			if(keysize > 2048) keysize = 2048;
			break;
		default:
			pgpErrString=sf_strdup("Unknown algorithm.");
			return -1;
	};

	/* SECRET Keyring */

	sprintf(buf, "/tmp/exaview-%s-%d-kis-XXXXXX",
		getformat("${object[0]}"), getpid());
	if((fd=mkstemp(buf)) == -1) { unlink(buf); return -1; };
	unlink(buf); close(fd);
	err = PGPNewFileSpecFromFullPath( context, buf, &secFileSpec );
	if(IsPGPError(err)) {
		PGPGetErrorString(err, sizeof(buf), buf);
		pgpErrString=sf_strdup(buf);
		return -1;
	};

	/* PUBLIC Keyring */
	sprintf(buf, "/tmp/exaview-%s-%d-kip-XXXXXX",
		getformat("${object[0]}"), getpid());
	if((fd=mkstemp(buf)) == -1) { unlink(buf); return -1; };
	unlink(buf); close(fd);
	err = PGPNewFileSpecFromFullPath( context, buf, &pubFileSpec );
	if(IsPGPError(err)) {
		PGPGetErrorString(err, sizeof(buf), buf);
		pgpErrString=sf_strdup(buf);
		return -1;
	};

	err = PGPOpenKeyRingPair ( context,
		kPGPKeyRingOpenFlags_Create |
		kPGPKeyRingOpenFlags_Mutable,
		secFileSpec, pubFileSpec,
		&keyringset );
	if(IsPGPError(err)) {
		PGPGetErrorString(err, sizeof(buf), buf);
		pgpErrString=sf_strdup(buf);
		return -1;
	};

	srandomdev();
	int i;
	unsigned int ent=PGPGetKeyEntropyNeeded( context,
		PGPOKeyGenParams( context, sigalg , keysize ),
		PGPOKeyGenFast( context, TRUE ),
		PGPOLastOption( context ));
	ent+=PGPGetKeyEntropyNeeded( context,
		PGPOKeyGenParams( context, encalg , keysize ),
		PGPOKeyGenFast( context, TRUE ),
		PGPOLastOption( context ));

	while(PGPGlobalRandomPoolGetEntropy() < ent)
		for(i=0; i<50; i++)
			PGPGlobalRandomPoolAddKeystroke(random());

	err = PGPGenerateKey( context, &masterkey,
		PGPOKeySetRef( context, keyringset ),
		PGPOKeyGenParams( context, sigalg, keysize ),
		PGPOKeyGenFast( context, TRUE ),
		PGPOKeyGenName( context, Info->from, strlen(Info->from) ),
		PGPOPassphrase( context, Info->pgp_passphrase ),
		PGPOExpiration( context, expirationDays ),
		PGPOLastOption( context )
	);

	if(IsPGPError(err)) {
		strcpy(buf, "Generate master key: ");
		int l=strlen(buf);
		PGPGetErrorString(err, sizeof(buf)-l, buf+l);
		pgpErrString=sf_strdup(buf);
		return -1;
	};

        err = PGPSetKeyAxiomatic( masterkey, PGPOLastOption( context));
	err = PGPSetDefaultPrivateKey( masterkey );
	if(IsPGPError(err)) {
		strcpy(buf, "Setting default private key: ");
		int l=strlen(buf);
		PGPGetErrorString(err, sizeof(buf)-l, buf+l);
		pgpErrString=sf_strdup(buf);
		return -1;
	};
//	PGPsdkSavePrefs( context );                                       

	if(encalg!=kPGPPublicKeyAlgorithm_RSA)
		if(keysize > 1024) keysize = 1024;

	err = PGPGenerateSubKey( context, &subkey,
		PGPOKeyGenMasterKey( context, masterkey ),
		PGPOKeyGenParams( context, encalg, keysize ),
		PGPOKeyGenFast( context, TRUE ),
		PGPOPassphrase( context, Info->pgp_passphrase ),
		PGPOExpiration( context, expirationDays ),
		PGPOLastOption( context )
	);

	err = PGPCommitKeyRingChanges( keyringset );
	pgp_RELOAD=1;

	if(IsPGPError(err)) {
		PGPGetErrorString(err, sizeof(buf), buf);
		pgpErrString=sf_strdup(buf);
		return -1;
	};

	err = PGPFreeKeySet( keyringset );

	return 0;

};

PGPError addKeyToKeySet( PGPKeyRef key, PGPKeySetRef keySet) {
    PGPError err;
    PGPKeySetRef tmpset;
    err = PGPNewSingletonKeySet( key, &tmpset );
    if(IsPGPError(err)) return err;
    err = PGPAddKeys( tmpset, keySet );
    PGPFreeKeySet( tmpset );
    return err;
};

PGPError addToWorkingKeySet(PGPKeySetRef keySet) {
    PGPKeyIterRef iter = NULL;
    PGPKeyRef key = 0;
    PGPKeyRef origkey = 0;
    PGPKeyID keyid;
    int newkeys = 0, newsigs = 0, newuids = 0, newrvks = 0;
    PGPKeyListRef keylist = 0;
    PGPKeySetRef origset = NULL;

    err = PGPNewEmptyKeySet( keyringset, &origset );
    err = PGPOrderKeySet( keySet, kPGPAnyOrdering, &keylist );

    err = PGPNewKeyIter( keylist, &iter );
    err = PGPKeyIterRewind( iter );
    err = PGPKeyIterNext( iter, &key);

    while( key ) {
        err = PGPGetKeyIDFromKey( key, &keyid );
		if(IsPGPError(err)) {
        	err = PGPKeyIterNext( iter, &key );
			if(IsPGPError(err)) break;
			continue;
		};
        err = PGPGetKeyByKeyID( keyringset, &keyid,
			kPGPPublicKeyAlgorithm_Invalid, &origkey );
        if(!IsPGPError(err))
			addKeyToKeySet( origkey, origset );
        err = PGPKeyIterNext( iter, &key );
		if(IsPGPError(err)) break;
    }

    err = PGPKeyIterRewind( iter );
	if(IsPGPError(err)) goto done;
    err = PGPKeyIterNext( iter, &key);
	if(IsPGPError(err)) goto done;

    while( key ) {

        err = PGPGetKeyIDFromKey( key, &keyid );
		if(IsPGPError(err)) {
        	err = PGPKeyIterNext( iter, &key );
        	if( IsPGPError(err) || key == 0 ) break;
			continue;
		};

        err = PGPGetKeyByKeyID( keyringset, &keyid,
                kPGPPublicKeyAlgorithm_Invalid,
                &origkey );

        if(IsPGPError(err)) {
            /* is a new key*/
			/*
            err = pgpShowKeyListFormat( filebPtr,
                    mainbPtr->workingRingSet, iter, key, kShow_Sigs |
                    kShow_Checks );
			*/

                newkeys++;
        }

		err = addKeyToKeySet( key, keyringset );
		if(IsPGPError(err)) break;

        err = PGPKeyIterNext( iter, &key );
        if( IsPGPError(err) || key == 0 ) break;

    };

    if(err == kPGPError_EndOfIteration) err = 0;

    if( newsigs == 0 && newkeys == 0 && newuids == 0 && newrvks == 0 ) {
		printf("No new keys or signatures found.\n");
        goto done;
    };

	printf("Found:\n");
    if (newkeys)
        printf("%4d new key(s)\n", newkeys);
    if (newsigs)
        printf("%4d new signatures(s)\n", newsigs);
    if (newuids)
        printf("%4d new user ID(s)\n", newuids);
    if (newrvks)
        printf("%4d new revocation(s)\n", newrvks);

    err = PGPPropagateTrust( keyringset );

done:
    if( iter )
        PGPFreeKeyIter( iter );
    if( keylist)
        PGPFreeKeyList( keylist );
    if( origset )
        PGPFreeKeySet( origset );

    return err;
};

PGPError pgpMyEventHandler(PGPContextRef context, struct PGPEvent *event, PGPUserValue userValue) {

	(void)context;
	err = kPGPError_NoErr;

	switch ( event->type ) {
	case kPGPEvent_FinalEvent:
		err = kPGPError_NoErr; break;
	case kPGPEvent_InitialEvent:
		err = kPGPError_NoErr; break;
	case kPGPEvent_BeginLexEvent:
		err = kPGPError_NoErr; break;
	case kPGPEvent_EndLexEvent:
		err = kPGPError_NoErr; break;
	case kPGPEvent_AnalyzeEvent:
		err = kPGPError_NoErr; break;
	case kPGPEvent_RecipientsEvent:
		err = kPGPError_NoErr; break;
	case kPGPEvent_PassphraseEvent:
		if(Info->pgp_passphrase) {
			if(pgpErrString)
				free(pgpErrString);
			pgpErrString = sf_strdup(cf("pgp-bad-passphrase"));
		};
		err = kPGPError_UserAbort; break;

	case kPGPEvent_KeyFoundEvent:
		err = addToWorkingKeySet((PGPKeySetRef)event->data.keyFoundData.keySet);
		break;

	case kPGPEvent_DetachedSignatureEvent:
		printf("DetachedSignatureEvent\n");
		err = kPGPError_NoErr; break;

	case kPGPEvent_SignatureEvent:
		if(userValue) {
			int *userData = (int *) userValue;
			*userData = 0;	// No signature in the file.

/*
			PGPGetKeyIDString(&event->data.signatureData.signingKeyID,
				kPGPKeyIDString_Abbreviated, buf);
			printf("KeyID: %s\n", buf);
*/

			if(event->data.signatureData.checked) {
				if(event->data.signatureData.verified) {
					*userData = 1;	// The signature verified
				} else {
					*userData = -1;	// Signature was bad
				};
			};
		};
		event->data.signatureData.signingKey = kInvalidPGPKeyRef;
		err = kPGPError_NoErr; break;

	case kPGPEvent_DecryptionEvent:
		err = kPGPError_NoErr; break;
/*
		printf("Symmetric cipher used: ");
		switch( event->data.decryptionData.cipherAlgorithm ) {
			case kPGPCipherAlgorithm_IDEA:
				printf("IDEA\n"); break;
			case kPGPCipherAlgorithm_3DES:
				printf("3DES\n"); break;
			case kPGPCipherAlgorithm_CAST5:
				printf("CAST5\n"); break;
			default:
				printf("unknown\n");
		};
		err = kPGPError_NoErr; break;
*/

	case kPGPEvent_OutputEvent:
		err = kPGPError_NoErr; break;
	case kPGPEvent_ErrorEvent:
		if(errno == ENOSPC)
			printf("Error: No space left on device.\n");
		err = kPGPError_NoErr; break;
	case kPGPEvent_WarningEvent:                                            
		err = kPGPError_NoErr; break;
	default:                                            
		err = kPGPError_NoErr; break;
	};

	return err;
};

int pgpimportkeys(char *keybuf) {
	if(!keybuf) return -1;
	PGPUInt32 nk1, nk2;

	if(pgpErrString) {
		free(pgpErrString);
		pgpErrString=NULL;
	};

	err = PGPCountKeys( keyringset, &nk1 );
	if(IsPGPError(err)) return -1;

	err = PGPDecode( context,
		PGPOInputBuffer( context, keybuf, strlen(keybuf)),
		PGPOAllocatedOutputBuffer( context, &outbuf, 2*1024*1024, &outbuflen ),
		PGPOKeySetRef( context, keyringset ),
		PGPOSendEventIfKeyFound( context, TRUE ),
		PGPOEventHandler( context, pgpMyEventHandler, NULL ),
		PGPOLastOption( context )
		);

	if(IsPGPError(err)) {
		strcpy(buf, "Decode error: ");
		int l = strlen(buf);
		PGPGetErrorString( err, sizeof(buf)-l, buf+l);
		pgpErrString=sf_strdup(buf);
		return -1;
	};

	err = PGPCountKeys( keyringset, &nk2 );
	if(IsPGPError(err)) return -1;

	if(PGPKeySetNeedsCommit(keyringset)) {
		PGPCommitKeyRingChanges(keyringset);
		pgp_RELOAD=1;
	};

	if(IsPGPError(err)) {
		strcpy(buf, "Add error: ");
		int l = strlen(buf);
		PGPGetErrorString( err, sizeof(buf)-l, buf+l);
		pgpErrString=sf_strdup(buf);
		return -1;
	};

	return 0;
};

char *pgpexportkeys(int exportprivate) {

	err = PGPExportKeySet( resultset,
		PGPOAllocatedOutputBuffer(context, &outbuf, 1024*1024, &outbuflen),
		PGPOExportPrivateKeys( context, exportprivate?TRUE:FALSE ),
		PGPOArmorOutput(context, TRUE ),
		PGPOVersionString(context, "Exaview version 4.0/pgp (c) 1998-2000 <vlm@spelio.net.ru>"),
		PGPOLastOption(context));
	if(IsPGPError(err)) return NULL;

	return (char *)outbuf;
};
