#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <string.h>
#include <errno.h>
#include <strfunc.h>

#include <common.h>
#include <mail.h>

CMail::CMail() {
	bodyp="";
	Flags=0;
	boundary_=NULL;
	bodysize_=-1;
	filesize_average_=-1;
	msgid_=-1;
	partid_=0;
	fname_=NULL;
	date_=0;
	msgnum_=0;

/*
** 1: filled by SetMail()
** 2: filled by SetBody()
** 3: filled by AddPart().
*/

	filled_=0;

	pcount_=psize_=0;
	parts_=NULL;
	freebodyatdestroy = 0;
	is_flags_=0;	// bit 'n+1' - flag, bit 'n' - it's availability.
};

#define ISFL_MIME_SET	1
#define ISFL_MIME		2

int
CMail::SetBody(char *body, int FatD = 0) {

	if(filled_ == 1 || filled_ == 2) {

		if(freebodyatdestroy)
			free(bodyp);

		bodysize_ = -1;

	} else if(filled_ || (bodysize_ != -1))
		return -1;

	freebodyatdestroy = FatD;
	bodyp = body;
	if(!bodyp)
		bodyp="";

	filled_ = 2;

	return 0;
};

char *
CMail::SetMail(char *mp, int msgnum=0, int msgid=-1, int partid=0) {

	if(filled_ || !mp || msgnum_)
		return NULL;

	char *ss=mp;

	msgnum_ = msgnum;
	msgid_ = msgid;
	partid_ = partid;

	while(*(++mp)) {
		if(*mp == '\r' && mp[1] == '\n') {
			*mp=0;
			mp+=2;
			header.AddHeadLine(ss);
			ss=mp;
			if(*mp != '\r' || mp[1] != '\n') continue;
			mp+=2;
			break;
		} else if(*mp == '\n') {
			mp++;
			if(*mp == ' ' || *mp == '\t') continue;
			*(mp-1)=0;
			header.AddHeadLine(ss);
			ss=mp;
			if(*mp != '\n') continue;
			mp+=1;
			break;
		} else continue;
	};

	filled_ = 1;	// Filled by SetMail();
	return bodyp=mp;
};

int
CMail::AddPart(CMail *m) {

	// mail has been already initialized by incompatible method
	if(filled_ > 0 && filled_ != 3)
		return NULL;

	if((pcount_ + 1) >= psize_) {
		psize_ = psize_ << 1;
		if(!psize_)
			psize_ = 4;
		parts_=(CMail **)sf_realloc(parts_, sizeof(CMail *) * psize_);
	}

	parts_[pcount_] = m;

	parts_[pcount_]->msgid_ = msgid_;

	parts_[pcount_]->partid_= partid_ * 100 + 1 + pcount_;

	pcount_++;

	filled_ = 3;
	return NULL;
};


int
CMail::Save(int fd) {
	size_t size;
	char *p;

	if(filled_ != 3) {
		p = header.Save(&size);
		write(fd, p, size);

		size = BodySize();
		if(size > 0)
			write(fd, bodyp, size);
		write(fd, "\n", sizeof("\n") - 1);
		return 0;
	};

	char *bnd = header.HeadLineParam("Content-Type:", "boundary");
	// Very strange. Then, recreating the Content-Type field.
	if(!bnd) {
		bnd = ConstructMIMEBoundary();
		char buf[128];
		sprintf(buf, "Content-Type: multipart/mixed; \n\tboundary=\"%s\"", bnd);
		header.AddHeadLine(buf);
	};
	size_t bndlen = strlen(bnd);

	p = header.Save(&size);
	write(fd, p, size);

	write(fd, "Please get a MIME-compliant mail reader.\n",
	sizeof("Please get a MIME-compliant mail reader.\n")-1);

	write(fd, "\n--", 3);
	write(fd, bnd, bndlen);
	write(fd, "\n", 1);

	for(int i=0; i < pcount_ ; i++) {
		if(i > 0) {
			write(fd, "\n--", 3);
			write(fd, bnd, bndlen);
			write(fd, "\n", 1);
		};
		parts_[i]->Save(fd);
	};

	write(fd, "\n--", 3);
	write(fd, bnd, bndlen);
	write(fd, "--\n\n", 4);

	return 0;
};


char *
CMail::Save() {
	size_t size;
	size_t n;
	char *svd;	// Pointer to the saved message
	char *p;	// Current position

	svd = header.Save(&size);
	p = svd + size;

	if(filled_ != 3) {
		n = size;

		size += BodySize() + 2;
		svd = (char *)sf_realloc(svd, size);

		p = svd + n;

		p += sprintf(p, "%s\n", bodyp);

		return svd;
	};

	char *bnd = header.HeadLineParam("Content-Type:", "boundary");

	// Very strange. Then, recreating the Content-Type field.
	if(!bnd) {
		bnd = ConstructMIMEBoundary();
		char buf[strlen(bnd) + 64];
		sprintf(buf, "Content-Type: multipart/mixed; \n\tboundary=\"%s\"", bnd);
		header.AddHeadLine(buf);
	};

	size_t bndlen = strlen(bnd);
	char *prt[pcount_];

	n = size;

	for(int i=0; i < pcount_ ; i++) {
		size += bndlen + 4;
		prt[i] = parts_[i]->Save();
		size += strlen(prt[i]);
	};
	size += (bndlen + 4) * 2 + 128;

	svd = (char *)sf_realloc(svd, size);
	p = svd + n;

	p += sprintf(p, "Please get a MIME-compliant mail reader.\n\n--%s\n", bnd);

	for(int i=0; i < pcount_ ; i++) {
		if(i > 0)
			p += sprintf(p, "\n--%s\n", bnd);
		p += sprintf(p, "%s", prt[i]);
	};

	p += sprintf(p, "\n--%s--\n", bnd);

	return svd;
};

int
CMail::Save(FILE *fd) {
	size_t size;
	char *p;

	if(filled_ != 3) {
		p = header.Save(&size);
		fwrite(p, size, 1, fd);

		size = BodySize();
		if(size > 0)
			fwrite(bodyp, size, 1, fd);

		fwrite("\n", sizeof("\n") - 1, 1, fd);

		return 0;
	}

	char *bnd = header.HeadLineParam("Content-Type:", "boundary");
	// Very strange. Then, recreating the Content-Type field.
	if(!bnd) {
		bnd = ConstructMIMEBoundary();
		char buf[128];
		sprintf(buf, "Content-Type: multipart/mixed; \n\tboundary=\"%s\"", bnd);
		header.AddHeadLine(buf);
	};
	size_t bndlen = strlen(bnd);

	p = header.Save(&size);
	fwrite(p, size, 1, fd);

	fwrite("Please get a MIME-compliant mail reader.\n",
	sizeof("Please get a MIME-compliant mail reader.\n")-1, 1, fd);

	fwrite("\n--", 3, 1, fd);
	fwrite(bnd, bndlen, 1, fd);
	fwrite("\n", 1, 1, fd);

	for(int i=0; i < pcount_ ; i++) {
		if(i > 0) {
			fwrite("\n--", 3, 1, fd);
			fwrite(bnd, bndlen, 1, fd);
			fwrite("\n", 1, 1, fd);
		};
		parts_[i]->Save(fd);
	};

	fwrite("\n--", 3, 1, fd);
	fwrite(bnd, bndlen, 1, fd);
	fwrite("--\n\n", 4, 1, fd);

	return 0;
};

CMail **
CMail::parts() {

	if(parts_) return parts_;

	if(!IsMultipart()) return NULL;

	// Determining parts boundary.
	char *bnd;
	char *tmp=(char *)HeadLineParam("Content-Type:", "boundary");

	if(!tmp)
		return NULL;

	// Allocate enough space to store "--" + boundary + '\0'
	size_t bndlen=strlen(tmp);
	bnd=(char *)sf_malloc(bndlen + 3);
	bnd[0]=bnd[1]='-';
	strcpy(bnd+2, tmp);
	bndlen+=2;

	psize_=4;
	pcount_=0;
	parts_=(CMail **)sf_malloc(psize_ * sizeof(CMail *));
	CMail *m;

	char *p=bodyp;
	int blen=BodySize();
	int ptid=partid_ * 100;

	if(msgid_ == -1)
		MsgId();

	p--;
	while(blen-- && *(++p)) {
		if(*p != '-' || p[1] != '-') continue;
		if(strncmp(bnd, p, bndlen)) continue;
			
		if(*(p+bndlen) == '\n' || *(p+bndlen) == '\r') {
			*p=0;
			p+=bndlen+1;
		} else if(!strncmp(p+bndlen, "--\n", 3)) {
			*(p-1)=0;
			p+=bndlen;
			break;
		} else if(!strncmp(p+bndlen, "--\r\n", 4)) {
			*(p-2)=0;
			p+=bndlen;
			break;
		} else {
			p+=bndlen;
			continue;
		};

		parts_[pcount_]=m=new CMail();

		tmp=m->SetMail(p, msgnum_, msgid_, ++ptid);
		m->is_flags_=ISFL_MIME | ISFL_MIME_SET;
		pcount_++;
		if(pcount_ >= psize_) {
			psize_ = psize_ << 1;
			parts_=(CMail **)sf_realloc(parts_, sizeof(CMail *) * psize_);
		};
		p=tmp-1;
	};

	free(bnd);

	if(!pcount_) {
		free(parts_);
		parts_=NULL;
	} else {
		filled_ = 3;
		parts_[pcount_]=0;
	};

	return parts_;
};

int
CMail::BodySize() {
	if(bodysize_ > 0)	// Not != -1, but exactly "> 0".
		return bodysize_;
	return bodysize_ = strlen(bodyp);
};

int
CMail::IsMime() {
	if(is_flags_ & ISFL_MIME_SET) return (is_flags_ & ISFL_MIME)?1:0;
	char *ver=HeadLineValue("MIME-Version:");
	if(!ver) {
		is_flags_ |= ISFL_MIME_SET;
		return 0;
	};
	if(!strcmp(ver, "1.0")) {
		is_flags_ |= ISFL_MIME_SET | ISFL_MIME;
		return 1;
	};

	char buf[strlen(ver)];
	char *p=buf;
	int incomment=0;

	ver--;
	while(*(++ver)) {
		if(*ver == '(' && !incomment) { incomment=1; continue; };
		if(incomment) {
			if(*ver == ')') incomment=0;
			continue;
		};
		if(*ver == ' ') continue;
		*p++=*ver;
	};
	*p=0;

	is_flags_ |= ISFL_MIME_SET;

	if(!strcmp(buf, "1.0"));
		is_flags_ |= ISFL_MIME;

	return (is_flags_ & ISFL_MIME)?1:0;
};

int
CMail::IsMultipart() {
	if(!IsMime()) return 0;
	char *mp=HeadLineParam("Content-Type:");
	if(!mp) return 0;
	if(!strncasecmp(mp, "multipart/", sizeof("multipart/")-1)) return 1;

	return 0;
};

struct recodefunc_ {
	char *name;
	char *(*func)(char *);
};

size_t
CMail::filesize_average() {

	if(filesize_average_ != -1) return filesize_average_;

	filesize_average_=BodySize();

	char *cte=HeadLineParam("Content-Transfer-Encoding:");
	if(!cte) cte="8bit";

	if(!strncasecmp(cte, "base64", 6)) {
		filesize_average_=filesize_average_/4*3;
	} else if(!strncasecmp(cte, "quoted-printable",
			sizeof("quoted-printable")-1)) {
		filesize_average_=filesize_average_ * 120 / 100;
	};

	return filesize_average_=(filesize_average_+1023)/1024;
};

recodefunc_ recodefunc[] = {
	{"windows-1251", windows_1251_to_koi8_r},
	{"iso-8859-5", iso_8859_5_to_koi8_r},
	{"cp-866", cp_866_to_koi8_r},
};

void *
CMail::file(size_t *size=NULL) {
	if(bodysize_ == -1) bodysize_=strlen(bodyp);

	if(!IsMime()) {
		if(size)
			*size=bodysize_;
			return bodyp;
	};

	void *decoded=bodyp;
	if(size) *size=bodysize_;

	char *cte=HeadLineParam("Content-Transfer-Encoding:");
	if(!cte) cte="8bit";

	if(!strncasecmp(cte, "7bit", 4) || !strncasecmp(cte, "8bit", 4)) {
		if(size) *size=bodysize_;
	} else if(!strncasecmp(cte, "base64", 6)) {
		decoded=base64_decode(bodyp, size);
	} else if(!strncasecmp(cte, "quoted-printable",
			sizeof("quoted-printable")-1)) {
		if(size) *size=bodysize_;
		decoded=quoted_printable_decode(bodyp);
	};

	char *ct=HeadLineParam("Content-Type:");
	if(!ct || strncasecmp(ct, "text/", 5)) return decoded;

	ct=HeadLineParam("Content-Type:", "charset");
	if(!ct) return decoded;

	int n=sizeof(recodefunc)/sizeof(recodefunc_);
	while(n--) {
		if(!strcasecmp(recodefunc[n].name, ct)) {
			decoded=recodefunc[n].func((char *)decoded);
			break;
		};
	};

	return decoded;
};

char *
CMail::filename() {
	if(fname_) return fname_;

	if(!IsMime())
		return fname_=sf_strdup("message.body");

	fname_=(char *)HeadLineParam("Content-Disposition:", "filename");
	if(!fname_) fname_=(char *)HeadLineParam("Content-Type:", "filename");
	if(!fname_) fname_=(char *)HeadLineParam("Content-Type:", "name");
	if(!fname_) return NULL;
	if(!*fname_) fname_="unnamed";

	return fname_=sf_strdup(fname_);
};

/*
int countcrc(char *str) {
    if(!str) return 0;

    register unsigned int nleft = strlen(str);
    register unsigned int *w = (unsigned int *)str;
    register int sum = 0;
    unsigned int answer = 0;
        
    while (nleft >= sizeof(int))  { sum += *w++; nleft -= sizeof(int); }
    if(nleft > 0) {
	if (nleft == 3) sum += answer;
	else if (nleft == 2) {
		*(unsigned char *)(&answer) = *(unsigned char *)w;
		sum += answer;
    	} else if (nleft == 1) {
		*(unsigned short *)(&answer) = *(unsigned short *)w;
		sum += answer;
	};
    };

    sum = (sum >> 24) + (sum & 0xffffff);
    sum += (sum >> 24);
    answer = sum & 0xffffff;

    return(answer);
};
*/

static int
counthash(char *str, unsigned int prevhash = 0) {
	register unsigned int h=prevhash;
	register const unsigned char *s=(unsigned char *)str;

	if(!str) return h;

	register int len=strlen(str);

	while(len--)
		h = h * 37 ^ (*s++ - ' ');
		h %= 1048583;

	if(h > 0x0fffffff) {
		h = (h >> 28) + (h & 0x0ffffff);
		h += (h >> 28);
	};

	return (h);
};

int
CMail::MsgId() {

	if(msgid_ != -1) return msgid_;

	char *mi=HeadLineValue("Message-ID:");

	msgid_ = counthash(mi);

	mi=HeadLineValue("From ");

	if(!mi) {
		mi=HeadLineValue("Date:");
		if(!mi)
			mi=bodyp;
	};

	return msgid_=counthash((char *)mi, msgid_);

};

int
CMail::PartId() {
	return partid_;
};

int
CMail::MsgNum() {
	return msgnum_;
};

time_t
CMail::Date() {

	if(date_)
		return date_;

	char *tmc;

	tmc=HeadLineValue("Date:");
	if(!tmc) tmc=HeadLineValue("From ");
	if(!tmc) return 0;

	date_=strtotime((char *)tmc);

	return date_;
};

CMail::~CMail() {
	if(filled_ == 2 && freebodyatdestroy)
		free(bodyp);
	if(boundary_)
		free(boundary_);
};

char *
CMail::HeadLineValue(char *a) {
	return header.HeadLineValue(a);
};

char *
CMail::HeadLineParam(char *a, char *b = NULL) {
	return header.HeadLineParam(a, b);
};

int
CMail::AddHeadLine(const char *field, const char *optParam = NULL) {

	if(!field) {
		errno = EINVAL;
		return -1;
	}

	int field_len = strlen(field);
	int optParam_len = optParam ? strlen(optParam) : 0;

	char *mem = (char *)sf_malloc(field_len + optParam_len + 2);

	memcpy(mem, field, field_len + 1);

	/*
	 * Append parameter.
	 */
	if(optParam) {
		mem[field_len] = ' ';
		memcpy(&mem[field_len + 1], optParam, optParam_len + 1);
	}

	return header.AddHeadLine(mem, 1);
};

int
CMail::i_CheckMIMEBoundary(const char *bnd) {
	char *p;

	if(!bnd)
		return -1;

	if(!filled_) {
		p=header.HeadLineParam("Content-Type:", "boundary");
		if(p && !strcmp(p, bnd))
			return -1;
		if(strlen(bnd) > 1)
			return 0;	// Any boundary will success.
		else
			return -1;
	};

	if(filled_ == 1 || filled_ == 2) {
		p=header.HeadLineParam("Content-Type:", "boundary");
		if(p && !strcmp(p, bnd))
			return -1;
		if(!bodyp || strstr(bodyp, bnd))
			return -1;
		return 0;
	};

	if(filled_ == 3) {
		if(!parts_)
			return 0;
		for(int i = 0; i < pcount_; i++)
			if(parts_[i]->i_CheckMIMEBoundary(bnd))
				return -1;
		return 0;
	};

	return 0;
};

char *
CMail::ConstructMIMEBoundary() {
	if(boundary_)
		return boundary_;

	if(!filled_) {	// Can't construct boundary
		return NULL;
	};

	char bnd[64];
	int bndn=0;

	while(1) {
		sprintf(bnd, "--Exaview-unique-boundary-%ld-%d_",
			(long)(time(NULL) - 960000000), bndn++);

		if(i_CheckMIMEBoundary(bnd))
				continue;

		break;
	};

	boundary_ = sf_strdup(bnd);

	return boundary_;
};

