#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <string.h>
#include <errno.h>
#include <mbox.h>
#include <mail.h>

CMailBox::CMailBox() {
	mboxfd=0;
	mboxf=NULL;
	mboxp=NULL;
	mails=NULL;
	mailcount=0;
};

CMailBox::~CMailBox() {
	if(mboxp) { munmap(mboxp, mboxsize+1); mboxp=NULL; }
	if(mboxfd) { close(mboxfd); mboxfd=0; mboxf=NULL; };
	if(mails) {
		while(mailcount--)
			delete(mails[mailcount]);
		free(mails);
		mails=NULL;
	};
};

int
CMailBox::open(char *fname) {
	if(mboxfd || mboxp) { errno = EALREADY; return -1; };
	if(!fname) { errno = EINVAL; return -1; };

	if(*fname != '/' || strstr(fname, "..")) {
		errno=EINVAL; return -1;
	};

	if((mboxfd=::open(fname, O_RDWR | O_SHLOCK)) == -1) return -1;

	char buf[sizeof("From ")];

	struct stat sb;
	if(fstat(mboxfd, &sb) || ((mboxsize=sb.st_size) < 8)) {
		close(mboxfd);
		mboxfd=0;
		return -1;
	};

	if(read(mboxfd, buf, 5) != 5 || strncmp(buf, "From ", 5)) {
		close(mboxfd); mboxfd=0;
		return -1;
	};

	mboxp=(char *)mmap(NULL, mboxsize+1, PROT_READ | PROT_WRITE, MAP_PRIVATE, mboxfd, 0);
	if(!mboxp) return -1;
	*(char *)(mboxp+mboxsize)='\0';

	if(fillpts()) return -1;

	mboxf = sf_strdup(fname);

	return 0;
};

int
CMailBox::fillpts() {

	char *point=mboxp;

	mlen=16;
	mails=(CMail **)malloc(sizeof(CMail *) * mlen);

	mails[0]=new CMail();
	point=mails[0]->SetMail(point, 1);
	char *lastp=mboxp+mboxsize;
	mailcount=1;

	char *fromsp="\nFrom ";
	char *tmp;

	while(point < lastp) {
		point++;
		if(*(int *)point != *(int *)fromsp) continue;
		if(strncmp(point, fromsp, 6)) continue;

		point++;
		mails[mailcount]=new CMail();
	tmp=mails[mailcount]->SetMail(point, mailcount+1);
		mailcount++;
		if(mailcount >= mlen) {
			mlen = mlen << 1;
			mails=(CMail **)realloc(mails, sizeof(CMail *) * mlen);
		};
		*(--point)=0;
		if(*(--point) == '\r') *point=0;
		point=tmp;
	};

	/* Additional <LF> or <CRLF> is the part of the RFC822, not user's data */

	*(--point)=0;
	if(*(--point) == '\r') *point=0;


	return 0;
};


int
CMailBox::count() { return mailcount; };

CMail *
CMailBox::getbynum(int msgnum) {
	if(msgnum < 1 || msgnum > mailcount) return NULL;
	return mails[msgnum-1];
};

CMail *
CMailBox::getbyid(int msgid) {
	if(mailcount < 1) return NULL;

	for(int i=0; i<mailcount; i++)
		if(mails[i]->MsgId() == msgid)
			return mails[i];

	return NULL;
};

int
CMailBox::rmbynum(int msgnum) {
	CMail *mail=getbynum(msgnum);
	if(!mail) return -1;

	delete(mail);

	msgnum--;
	for(int i=msgnum; i<mailcount-1; i++)
		mails[i]=mails[i+1];

	mailcount--;
	return 0;
};

int
CMailBox::rmbyid(int msgid) {
	CMail *mail=NULL;
	int i;

	for(i=0; i < mailcount; i++)
		if(mails[i]->MsgId() == msgid) {
			mail=mails[i];
			break;
		};

	if(!mail) return -1;
	delete(mail);

	for(; i<mailcount-1; i++)
		(mails[i]=mails[i+1])->msgnum_--;

	mailcount--;
	return 0;
};

static sigset_t oldsig;
sigset_t bls;

void
blocksig() {
        sigemptyset(&bls);
        sigaddset(&bls, SIGHUP);
        sigaddset(&bls, SIGINT);
        sigaddset(&bls, SIGQUIT);
        sigaddset(&bls, SIGTERM);
        sigprocmask(SIG_BLOCK, &bls, &oldsig);
};

void
unblocksig() {
        sigprocmask(SIG_SETMASK, &oldsig, NULL);
};

int
CMailBox::Save() {
	int tfd;

	if(!mboxfd) return -1;

	char *tmpfn=sf_strdup("/tmp/vread.delete.XXXXXX");

	if((tfd=mkstemp(tmpfn)) == -1) {
		free(tmpfn);
		return -1;
	};
	unlink(tmpfn);
	free(tmpfn);

	for(int i=0; i<mailcount; i++) {
		mails[i]->Save(tfd);
	};

	if(flock(mboxfd, LOCK_EX)) {
		close(tfd);
		return -1;
	};

	if(lseek(tfd, 0L, SEEK_SET) == -1) return -1;
	if(lseek(mboxfd, 0L, SEEK_SET) == -1) return -1;

	char buf[4096];
	size_t wrlen;
	struct stat sb;

	if(fstat(tfd, &sb)) { close(tfd); return -1; };

	/* We are praying, don't crash now!!! */
	blocksig();

	while((wrlen=read(tfd, buf, sizeof(buf))) > 0)
		write(mboxfd, buf, wrlen);

	close(tfd);

	/*
	** According to the actual things, ftruncate definitely
	** can't fail at this moment
	*/
	ftruncate(mboxfd, sb.st_size);

	fsync(mboxfd);

	unblocksig();

	if(flock(mboxfd, LOCK_SH))
		return 0;

	return 0;
};

struct valmail_ {
	char *value;
	CMail *mail;
};

int
sortcmp(const void *pa, const void *pb) {
	register valmail_ *a = (valmail_ *)pa;
	register valmail_ *b = (valmail_ *)pb;

	if(!a->value && !b->value) return 0;
	if(!a->value) return -1;
	else if(!b->value) return 1;

	return strcasecmp(a->value, b->value);
	return 0;
};

int
CMailBox::sort(char *f) {
	if(!f) return -1;
	if(mailcount < 2) return 0;

	valmail_ vm[mailcount];

	char *f2=(char *)malloc(strlen(f)+2);
	sprintf(f2, "%s:", *f=='-'?f+1:f);

	int n=mailcount;

	if(strcasecmp(f2, "date:")) {
		while(n--)
			vm[n].value=(vm[n].mail=mails[n])->HeadLineValue(f2);
	} else {
		time_t dt;
		char buf[16];

		while(n--) {
			dt=(vm[n].mail=mails[n])->Date();
			sprintf(buf, "%010d", (int)dt);
			vm[n].value=sf_strdup(buf);
		};
	};

	free(f2);

	qsort(vm, mailcount, sizeof(valmail_), sortcmp);

	if(*f == '-') {
		n=mailcount;
		while(n--) {
			mails[n]=vm[mailcount-n-1].mail;
			mails[n]->msgnum_=n+1;
		};
	} else {
		n=mailcount;
		while(n--) {
			mails[n]=vm[n].mail;
			mails[n]->msgnum_=n+1;
		};
	};

	return 0;
};

time_t
CMailBox::lastmodified() {
	struct stat sb;

	if(!mboxfd) return (time_t)0;

	if(fstat(mboxfd, &sb)) return (time_t)0;

	return sb.st_ctime+1;
};

