#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strfunc.h>
#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>
#include <mbox.h>
#include <mail.h>
#include <lgba.h>

#include <common.h>
#include "main.h"

#ifdef PGP

#include <lpgp.h>

#define FLAG_PGP_DECRYPTED	0x1
#define FLAG_PGP_SIGNED		0x2
#define FLAG_PGP_SIGNOK		0x4

#endif


char *cf(char *a) {
	register char *z=cfgget(a);
	static char empty[] = "";
	return z?z:empty;
};

char *htmladdr(char *);
void preparefmt();
void display();
void remove();

fmt_base *fb=NULL;

char **c2userinfo(char *, void *);
char **c2mailinfo(char *, void *);
char **c2mboxinfo(char *, void *);

class CMail *cMessage = NULL;
class CMailBox *cMailBox = NULL;

int
main() {
	char *p;
	int forcerefresh=0;

	// Protection. Give the LDAP rights, but no more.
	setgid(84);

	init();

	forcerefresh |= param("remove") || param("sortby");

	cMailBox = new CMailBox();

	{
		char *mbl = cfgget("mailboxdir");

		if(!mbl || *mbl != '/')
			mbl="/var/mail";

		char *fbuf = (char *)alloca(
			strlen(mbl) + strlen(Info->username) + 2);
		p = fbuf + sprintf(fbuf, "%s/%s", mbl, Info->username);

		if(cMailBox->open(fbuf)) {
			printf("Expires: %s\n",
				timetostr(0, TFMT_RFC822 | TFMT_OLD822GMT));
			printf("Content-type: text/html");
			if((p=cfgget("charset")))
					printf("; charset=%s", p);
			printf("\n\n");
			printf("%s", cf("start"));
			if(Info->msgid == -1)
				printf("%s", cf("no_messages"));
			else
				printf("%s", cf("not_found"));
			printf("%s", cf("end"));
			return 2;
		}
	}

	if(!forcerefresh && (p=getenv("HTTP_IF_MODIFIED_SINCE"))) {
		if(strtotime(p) > cMailBox->lastmodified()) {
			printf(
			"Expires: %s\n"
			"Status: 304 Mailbox intact, use cached page.\n\n"
				,
				timetostr(0, TFMT_RFC822 | TFMT_OLD822GMT));
			exit(2);
		}
	}

	if(forcerefresh) {
		printf("Expires: %s\n",
			timetostr(0, TFMT_RFC822 | TFMT_OLD822GMT));
	}

	if(Info->sortby && *Info->sortby)
		cMailBox->sort(Info->sortby);

	/*
	 * Remove deleted letters.
	 */
	if(param("remove")) {
		remove();
	}

	if(!cMailBox->count()
			|| (Info->msgid != -1
				&& !cMailBox->getbyid(Info->msgid))) {
		// Initialize html, we did not made it earlier
		printf("Content-type: text/html");
		if((p=cfgget("charset")))
			printf("; charset=%s", p);
		printf("\n\n%s", cf("start"));

		if(!cMailBox->count())
			printf("%s", cf("no_messages"));
		else
			printf("%s", cf("not_found"));

		printf("%s", cf("end"));
		return 0;
	}

	if(Info->msgid != -1 && Info->path_info) {
		cMessage = cMailBox->getbyid(Info->msgid);
		if(cMessage) {
			if(findpart(cMessage,
					Info->path_info, Info->partid) >= 0) {
				/*
				 * We found part and displayed it.
				 * Mission accomplished.
				 */
				return 0;
			}
		}
		printf("Content-type: text/html");
		if((p=cfgget("charset")))
			printf("; charset=%s", p);
		printf("\n\n%s", cf("start"));
		printf("%s", cf("not_found"));
		printf("%s", cf("end"));
		return 0;
	}

	printf("Content-type: text/html");
	if((p=cfgget("charset")))
		printf("; charset=%s", p);
	printf("\n\n%s", cf("start"));


	preparefmt();

	fflush(stdout);
	if(Info->msgid == -1) {
		p=formatf(fb, cf("list_start"), NULL);
		fwrite(p, format_lastsize(fb), 1, stdout);

		for(int num = 1; num <= cMailBox->count(); num++) {
			cMessage = cMailBox->getbynum(num);
			p=formatf(fb, cf("list_mail"), NULL);
			fwrite(p, format_lastsize(fb), 1, stdout);
		};

		p=formatf(fb, cf("list_end"), NULL);
		fwrite(p, format_lastsize(fb), 1, stdout);
	} else {
		if(Info->msgnum < 1)
			Info->msgnum = cMailBox->getbyid(Info->msgid)->MsgNum();
		cMessage = cMailBox->getbynum(Info->msgnum);
		p=formatf(fb, cf("display_start"), NULL);
		fwrite(p, format_lastsize(fb), 1, stdout);
			display();
		p=formatf(fb, cf("display_end"), NULL);
		fwrite(p, format_lastsize(fb), 1, stdout);
	}

	printf("%s", cf("end"));

	return 0;
}



void
preparefmt() {

	if(!fb) {
		if(!(fb=format_init())) return;
		format_metarule(fb, '{', '}', c2userinfo);
		format_metarule(fb, '`', '\'', c2mailinfo);
		format_metarule(fb, '(', ')', c2mboxinfo);
	};

};

char **c2userinfo(char *found, void *) {
	return sarray(getattr2(found), 0);
}

char *
_urlenc_handler(svect *sl) {

	if(sl->count < 3) {
		sclear(sl);
		return NULL;
	}

	return url_encode(sl->list[2]);
}

char **c2mailinfo(char *fld, void *) {
	static char *c2[]={NULL, NULL};
	static char buf[64];
	char *tmp;

	if(!*fld)
		return NULL;

	if(*fld == 'B') {
		if(!(*c2=(char *)cMessage->file()))
			return NULL;
		*c2=mime_word_decode(*c2);
		*c2=html_quote(*c2);
		int width=atoi(fld+1);
		if(width > 0)
		*c2=limittextwidth(*c2, width);
		return c2;
	}

	if(*fld == 'b') {
		static sed_t *se;
		static int se_initialized = 0;

		if(!(*c2=(char *)cMessage->file()))
			return NULL;

		if(se == NULL && se_initialized++ == 0) {
			char *expr;

			expr = (char *)alloca(128
			+ (Info->script_name?strlen(Info->script_name):0));

			sprintf(expr,
				"s/([[:space:]]SRC=[\"])?(cid:[^ \"\r\n>]+)/"
				"\\1%s\\/\\@\\?msgid=%d"
				"/gei",
				Info->script_name?Info->script_name:"eread.cgi",
				Info->msgid
			);

			se = sed_compile(expr);
			if(se) {
				sed_handler(se, _urlenc_handler);
			}
		}

		if(se) {
			char *p = sed_exec(se, *c2);
			if(p) *c2 = p;
		}

		return c2;
	}

	if(fld[1]) {
		/*
		 * if long
		 */

		tmp = cMessage->HeadLineValue(fld);

		/* Empty field and the absense of this field are equal */
		if(!tmp || !strlen(tmp)) return NULL;

		*c2 = htmladdr(tmp);

		return c2;
	}

	// Content-Description (Caption).
	if(*fld == 'C') {
		if(!(tmp=cMessage->HeadLineValue("Content-Description:"))
			|| !*tmp)
			return NULL;
		tmp = mime_word_decode(tmp);
		tmp = html_quote(tmp);
		*c2=tmp;
		return c2;
	}

	// Content-Description (Caption), quoted.
	if(*fld == 'c') {
		if(!(tmp=cMessage->HeadLineValue("Content-Description:")) || !*tmp)
			return NULL;
		tmp = mime_word_decode(tmp);
		tmp = html_quote(tmp);
		*c2=tmp;
		return c2;
	}


	// Is message PGP-encoded?
	if(*fld == 'P') {
#ifdef PGP
		static slist *sl=NULL;
		if(!sl) sl=sinit();
		else sclear(sl);

		if(cMessage->Flags & FLAG_PGP_SIGNED) {
			sadd(sl, cf("pgp-message-signed"));
			if(cMessage->Flags & FLAG_PGP_SIGNOK) {
				sadd(sl, cf("pgp-message-signature-good"));
			} else {
				sadd(sl, cf("pgp-message-signature-bad"));
			}
		}

		if(cMessage->Flags & FLAG_PGP_DECRYPTED)
			sadd(sl, formatf(fb, cf("pgp-message-decoded"), NULL));

		if(sl->count)
			return sarray(sl, 0);

		if((tmp=cMessage->HeadLineParam("Content-Type:", NULL))) {
			if(!strcmp(tmp, "application/octet-stream"))
				return NULL;
		}

		size_t size;
		char *file=(char *)cMessage->file(&size);
		if(!file || size < 100) return NULL;
		for(int i=0; i<200 && *file; i++, file++) {
			if(*file != '-') continue;
			if(!strncmp(file, "-----BEGIN PGP", 14)) {
				sadd(sl, cf("pgp-message-encoded"));
				sadd(sl, formatf(fb, cf("pgp_decode"), NULL));
				return sarray(sl, 0);
			}
		}
#else
		return NULL;
#endif
	}

	// Shorten and localize the date.
	if(*fld == 'D') {
		struct tm *tm;
		time_t dt=cMessage->Date();

		tm=localtime(&dt);
		strftime(buf, sizeof(buf), "%d&nbsp;%b,&nbsp;%H:%M", tm);
		*c2=buf;
		return c2;
	}

	// Subject field with additional checks.
	if(*fld == 'S') {
		if(!(tmp=cMessage->HeadLineValue("Subject:")) || !*tmp)
			return NULL;
		tmp=mime_word_decode(tmp);
		tmp=html_quote(tmp);
		*c2=tmp;
		return c2;
	}

	// Average file size
	if(*fld == 'A') {
		int z=cMessage->filesize_average();
		if(z > 3*1024) snprintf(buf, sizeof(buf), "%dM", z >> 20);
		else snprintf(buf, sizeof(buf), "%dK", z);
		*c2=buf;
		return c2;
	}

	// Length of the body
	if(*fld == 'L') {
		int z=cMessage->BodySize();
		if(z > 3*1024*1024) snprintf(buf, sizeof(buf), "%dM", z >> 20);
		else if(z > 3000) snprintf(buf, sizeof(buf), "%dK", z >> 10);
		else snprintf(buf, sizeof(buf), "%d", z);
		*c2=buf;
		return c2;
	}

	// Filename, if present
	if(*fld == 'F') {
		char *f=cMessage->filename();
		if(!f) return NULL;
		f = mime_word_decode(f);
		f = html_quote(f);
		*c2=f;
		return c2;
	}

	// Filename, if present, with url_encode
	if(*fld == 'f') {
		char *f=cMessage->filename();
		if(!f) f="untitled";
		*c2 = url_encode(f);
		return c2;
	}

	// Is this a multipart message?
	if(*fld == 'a') {
		if(!cMessage->IsMultipart())
			return NULL;
		*c2="";
		return c2;
	}

	if(*fld == 'm') {
		*c2 = genstr(cMessage->MsgId());
		return c2;
	}

	if(*fld == 'p') {
		*c2 = genstr(cMessage->PartId());
		return c2;
	}

	return NULL;

}

char **c2mboxinfo(char *what, void *) {
	static char *c2[]={NULL, NULL};
	int n;
	switch(*what) {
		case 'c':
			*c2 = genstr(cMessage->MsgNum());
			return c2;
		case 'p':
			n=cMessage->MsgNum() - 1;
			if(n < 1) return NULL;
			*c2 = genstr( cMailBox->getbynum(n)->MsgId() );
			return c2;
		case 'n':
			n=cMessage->MsgNum() + 1;
			if(n > cMailBox->count()) return NULL;
			*c2 = genstr( cMailBox->getbynum(n)->MsgId() );
			return c2;
		case 'q':
			*c2 = genstr( cMailBox->count() );
			return c2;
	};
	return NULL;
}


char *
_htmladdr_sed_cb(svect *sv) {
	if(cMessage) {
		static char buf[32];
		int msgid = cMessage->MsgId();
		sprintf(buf, "reply=%d&", msgid);
		return buf;
	} else {
		return "";
	}
}

char *
htmladdr(char *addr) {
	static sed_t *email_expr = NULL;
	static int init=0;

	addr = mime_word_decode(addr);
	addr = html_quote(addr);

	if(init == -1)
		/* Tried to compile expression, but failed. */
		return addr;

	if(init == 0) {
		char *expr =
			"s/([_a-z0-9.-]+@[a-z0-9.-]+)/"
			"<A HREF=\"ewrite.cgi?\\@to=\\1\">\\1<\\/A>"
			"/gei";
		email_expr = sed_compile(expr);
		if(email_expr) {
			sed_handler(email_expr, _htmladdr_sed_cb);
			init = 1;
		} else {
			init = -1;
			return addr;
		}
	}

	char *r = sed_exec(email_expr, addr);
	return r?r:addr;

}

void
display() {
	char *ct = cMessage->HeadLineParam("Content-Type:");

	if(ct) {
		char *p = (char *)alloca(strlen(ct) + 1);
		strcpy(p, ct);
		ct = p;
		/* "text/html; charset=xxx" => "text/html" */
		if((p = strchr(ct, ';')))
			*p = '\0';
	} else {
		ct = "text/plain";
	}

	char *f=NULL;

	if(!cMessage->IsMime()) {
		if(!strcasecmp(ct, "text/html"))
			f=formatf(fb, cf("text/html"), NULL);
		else
			f=formatf(fb, cf("plain"), NULL);
		
		fwrite(f, format_lastsize(fb), 1, stdout);
		return;
	}

	if(!(cMessage->IsMultipart())) {
		char *tmp;

		// This is the common case with the attached files.

		if((tmp=cMessage->HeadLineParam("Content-Disposition:")) &&
			!strcmp(tmp, "attachment")) {
			f=formatf(fb, cf("file"), NULL);
			fwrite(f, format_lastsize(fb), 1, stdout);
			return;
		};

		if(!strncasecmp(ct, "text/", sizeof("text/") - 1)) {
			if(!strcasecmp(ct, "text/html")) {
				f=formatf(fb, cf("text/html"), NULL);
			} else {
				f=formatf(fb, cf("text/plain"), NULL);
			}
			fwrite(f, format_lastsize(fb), 1, stdout);
			return;
		}

		if(!strncasecmp(ct, "image/", sizeof("image/")-1)) {
			f=formatf(fb, cf("image"), NULL);
			fwrite(f, format_lastsize(fb), 1, stdout);
			return;
		}

		f=formatf(fb, cf("file"), NULL);
		fwrite(f, format_lastsize(fb), 1, stdout);

		return;
	}


/***********************************************************/
/***                   Mail parts                        ***/
/***********************************************************/

	CMail **mails=NULL;

#ifdef PGP

	/*
	**	MULTIPART/SIGNED
	**
	**	protocol = application/pgp-signature
	*/

	if(!strcasecmp(ct, "multipart/signed")
	&& (f=cMessage->HeadLineParam("Content-Type:", "protocol"))
	&& !strcasecmp(f, "application/pgp-signature") ) {

		if(pgpinit("")) {
			printf("%s", cf("pgp-init-failed"));
			goto raf;
		};

		int result=0;
		char *msg;

		msg = cMessage->Save();
		msg = (char *)pgpdecode(msg, NULL, NULL, &result);
		pgpcleanup();

		if(msg) {
			CMail *mail = new CMail();
			mail->SetMail(msg, cMessage->MsgNum(),
				cMessage->MsgId(), cMessage->PartId());
			mail->AddHeadLine("MIME-Version: 1.0");

			mail->Flags |= FLAG_PGP_DECRYPTED;
			mail->Flags |= FLAG_PGP_SIGNED;
			if(result == 1)
				mail->Flags |= FLAG_PGP_SIGNOK;

			cMessage = mail;
			display();
			return;
		};

		goto raf;
	}

	/*
	**	MULTIPART/ENCRYPTED
	**
	**	protocol = application/pgp-encrypted
	*/

	if(!strcasecmp(ct, "multipart/encrypted")
	&& (f=cMessage->HeadLineParam("Content-Type:", "protocol"))
	&& !strcasecmp(f, "application/pgp-encrypted") ) {

		if(pgpinit("")) {
			printf("%s", cf("pgp-init-failed"));
			goto raf;
		};

		char *msg;
		int result;

		msg = (char *)cMessage->Save();
		msg = (char *)pgpdecode(msg, NULL, NULL, &result);
		pgpcleanup();

		if(Info->pgp_passphrase && !msg)
			printf("%s\n", pgpErrString);

		if(!msg) {
			formatf(fb, cf("pgp_decode_whole"), NULL);
			fwrite(format_lastresult(fb), format_lastsize(fb), 1, stdout);
		};

		if(msg) {
			CMail *mail = new CMail();
			mail->SetMail(msg, cMessage->MsgNum(),
				cMessage->MsgId(), cMessage->PartId());
			mail->AddHeadLine("MIME-Version: 1.0");
			mail->Flags |= FLAG_PGP_DECRYPTED;

			if(result != -1)
				mail->Flags |= FLAG_PGP_SIGNED;
			if(result == 1)
				mail->Flags |= FLAG_PGP_SIGNOK;

			cMessage = mail;
			display();
			return;
		};

		goto raf;
	};

raf:

#endif	/* PGP */

	mails = cMessage->parts();
	if(!mails) {
		f=formatf(fb, cf("text/plain"), NULL);
		fwrite(f, format_lastsize(fb), 1, stdout);
		return;
	} else {
		// Flags protection
		for(CMail **mls=mails; *mls; mls++) {
#ifdef PGP
			if(cMessage->Flags & FLAG_PGP_DECRYPTED)
				(*mls)->Flags |= FLAG_PGP_DECRYPTED;
			if(cMessage->Flags & FLAG_PGP_SIGNED)
				(*mls)->Flags |= FLAG_PGP_SIGNED;
			if(cMessage->Flags & FLAG_PGP_SIGNOK)
				(*mls)->Flags |= FLAG_PGP_SIGNOK;
#endif	/* PGP */
		};
	};

	/*
	**	MULTIPART/ALTERNATIVE
	*/

	if(!strcasecmp(ct, "multipart/alternative") &&
		mails[0] && mails[1] && !mails[2]) {
		int t1=0, t2=0;

		char *ct1=mails[0]->HeadLineParam("Content-Type:");
		if(!ct1) goto siq;
		if(!strncasecmp(ct1, "text/plain", sizeof("text/plain")-1))
			t1=1;
		else if(!strncasecmp(ct1, "text/html", sizeof("text/html")-1))
			t1=2;

		char *ct2=mails[1]->HeadLineParam("Content-Type:");
		if(!ct2) goto siq;
		if(!strncasecmp(ct2, "text/plain", sizeof("text/plain")-1))
			t2=1;
		else if(!strncasecmp(ct2, "text/html", sizeof("text/html")-1))
			t2=2;

		/*
		 * Display only one html message.
		 */
		if((t1 + t2) == 3) {

			if(t1 == 1) {
				cMessage=mails[1];
				display();
				return;
			}

			if(t1 == 2) {
				cMessage=mails[0];
				display();
				return;
			}

		}

	}

	/*
	**	MULTIPART/RELATED
	*/

	if(!strcasecmp(ct, "multipart/related")) {
		char *ct1=mails[0]->HeadLineParam("Content-Type:");
		if(!ct1) goto siq;
		if(strncasecmp(ct1, "text/html", sizeof("text/html") - 1) &&
			strncasecmp(ct1, "multipart/", sizeof("multipart/")-1))
				goto siq;
		cMessage=mails[0];
		display();
		return;
	}


siq:

	for(CMail **mls=mails; (cMessage=*mls); mls++) {
		/*
		 * Display subsequent parts of the message.
		 */
		display();
	}

	return;
}

void
remove() {
	int mi; 
	int rmd=0;

	slist *sl=param2("letter", 0);

	if(!sl || !sl->count) return;

	for(int i=sl->count-1; i>=0; i--) {
		mi=atoi(sl->list[i]);
		if(cMailBox->rmbyid(mi) == 0)
			rmd++;
	};

	if(!rmd) return;

	cMailBox->Save();
}



