#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strfunc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>
#include <lgba.h>

#include <common.h>
#include <mbox.h>
#include <mail.h>

char *cf(char *a) {
	register char *z=cfgget(a);
	static char empty[] = "";
	return z ? z : empty;
}

char *fullname="unknown";
char *to=NULL;
char *cc=NULL;
char *bcc=NULL;
char *subj=NULL;
char *body=NULL;
char *sign="";
int dosend=0;
int reply=-1;

int sendmessage();
char *getreplybody(CMail &);

CMailBox *cMailBox=NULL;
CMail *cMessage=NULL;

char **c2userinfo(char *, void *);
char **c2mailinfo(char *, void *);
char **c2mboxinfo(char *, void *);

int
main() {
	char *p;

	init();

	if(param("send") || param("to"))
		printf("Expires: %s\n",
			timetostr(0, TFMT_RFC822 | TFMT_OLD822GMT));

	printf("Content-Type: text/html");
	if((p=cfgget("charset")))
		printf("; charset=%s", p);
	printf("\n\n%s", cf("start"));

	to=param("to");
	cc=param("cc");
	bcc=param("bcc");
	body=param("body");
	subj=param("subject");
	if(param("send")) dosend=1;

#ifdef NO_LGBA
	fullname=Info->username;
	sign=(char *)sf_malloc(strlen(fullname) + 10);
	sprintf(sign, "\n\n-- \n%s", fullname);
#else
	fullname=Info->fullname;
	if(getattr2("email-signature")) {
		sign=sf_strdup(getformat("\n\n${email_signature+\n}\n"));
	} else {
		if((p=cfgget("domain"))) {
			char *tmp=(char *)sf_malloc(64 + strlen(p));
			sprintf(tmp, "\n\n-- \n${cn[0]}\n${%s[0]}@%s\n",
				lgba_ObjectAttribute(), p );
			sign=sf_strdup(getformat(tmp));
			free(tmp);
		} else {
			sign=sf_strdup(getformat("\n\n-- \n${cn[0]}\n"));
		};
	};
	if(!sign) {
		fprintf(stderr, "Memory allocation error.\n");
		printf("Memory allocation error.\n");
		exit(2);
	};
#endif

	if(!body && (p=param("reply")))
		reply=atoi(p);
	if(reply != -1) {
		char *mbl=cfgget("mailboxdir");
		cMailBox = new CMailBox();

        if(!mbl || *mbl != '/') mbl="/var/mail";
        char *fbuf=(char *)sf_malloc(strlen(Info->username) + strlen(mbl) + 2);
        p=fbuf+sprintf(fbuf, "%s/%s", mbl, Info->username);
        while(*p) *p++=tolower(*p);

        if(cMailBox->open(fbuf) == 0)
			cMessage=cMailBox->getbyid(reply);

		if(cMessage) {
			slist *sl;
			char *p = getreplybody(*cMessage);
			if(!p) p = "";
			p = limittextwidth(p, 70);
			sl = split(p, "\r\n", 1);
			p = sjoin(sl, "\n> ");
			body=(char *)sf_malloc(strlen(p) + strlen(sign?sign:"") + 10);
			sprintf(body, "> %s\n%s", p, sign?sign:"");

			if(!(p=subj=cMessage->HeadLineValue("Subject:")))
				subj="[no subject]";

			if(strncmp(subj, "Re: ", 4)) {
				subj=(char *)sf_malloc(strlen(subj) + 5);
				sprintf(subj, "Re: %s", p);
			} else {
				subj=sf_strdup(subj);
			};

			if(!to) {
				to = cMessage->HeadLineValue("Reply-To:");
				if(!to)
				to = cMessage->HeadLineValue("From:");
				if(to) to=sf_strdup(to);
				cc = cMessage->HeadLineValue("Cc:");
				if(cc) cc=sf_strdup(cc);
			};
		} else {
			body=(char *)sf_malloc(strlen(sign)+128);
			sprintf(body, "[Can't get source email for reply]\n\n%s", sign);
		};

        free(fbuf);
	};

	if(!body)
		body = sign;
	if(!subj)
		subj = "";

	if(dosend) {
		if(sendmessage() == 0)
			printf("%s", cf("mail-sent"));
		else {
			printf("%s\n", cf("mail-not-sent"));
		};
	} else {
		fmt_base *fb;
		fb=format_init();
		format_metarule(fb, '{', '}', c2userinfo);
		format_metarule(fb, '`', '\'', c2mailinfo);
		format_metarule(fb, '(', ')', c2mboxinfo);
		printf("%s", formatf(fb, cf("write-form"), NULL));
	};

	printf("%s", cf("end"));

	return 0;
};

char **
c2userinfo(char *field, void *) {
	return sarray(getattr2(field), 0);
};

char **c2mboxinfo(char *what, void *) {
	static char *c2[]={NULL, NULL};
	int n;

	if(!cMailBox) return NULL;

	switch(*what) {
		case 'c':
			*c2 = genstr(cMessage->MsgNum());
			return c2;
		case 'p':
			n=cMessage->MsgNum() - 1;
			if(n < 1) return NULL;
			*c2 = genstr( cMailBox->getbynum(n)->MsgId() );
			return c2;
		case 'n':
			n=cMessage->MsgNum() + 1;
			if(n > cMailBox->count()) return NULL;
			*c2 = genstr( cMailBox->getbynum(n)->MsgId() );
			return c2;
		case 'q':
			*c2 = genstr( cMailBox->count() );
			return c2;
	};
	return NULL;
};

char **
c2mailinfo(char *field, void *) {
	static char *c2[]={NULL, NULL};

	if(!*field) return NULL;

	if(!strcmp(field, "to")) {
		if(!to) return NULL;
		*c2=html_quote(to);
		return c2;
	};

	if(!strcmp(field, "cc")) {
		if(!cc) return NULL;
		*c2=html_quote(cc);
		return c2;
	};

	if(!strcmp(field, "from")) {
		if(!fullname) return NULL;
		*c2=html_quote(fullname);
		return c2;
	};

	if(!strcmp(field, "bcc")) {
		if(!bcc) return NULL;
		*c2=html_quote(bcc);
		return c2;
	};

	if(!strcmp(field, "subject")) {
		if(!subj) return NULL;
		*c2=html_quote(subj);
		return c2;
	};

	if(!strcmp(field, "body")) {
		if(!body) return NULL;
		*c2=html_quote(body);
		return c2;
	};

	return NULL;
};


char *
getreplybody(CMail &mail) {
	char *p;

	if(!mail.IsMime()) return (char *)mail.file();

	if(!(mail.IsMultipart())) {
		p=mail.HeadLineParam("Content-Type:");
		if(p && !strcasecmp(p, "text/plain"))
			return (char *)mail.file(NULL);
		return NULL;
	};

	CMail **mails=mail.parts();

	if(!mails) return (char *)mail.file(NULL);

	CMail **mls=mails-1;

	while(*(++mls))
		if((p=getreplybody(**mls)))
				return p;

	return NULL;
};

