#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strfunc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <ctype.h>
#include <lgba.h>

#include <common.h>
#include <mbox.h>
#include <mail.h>

#ifdef PGP
#include <lpgp.h>
#endif

char *
cf(char *);

extern char *fullname;
extern char *to;
extern char *cc;
extern char *bcc;
extern char *subj;
extern char *body;
extern char *sign;

FILE *openmta();

char *getmimetype(char *fname);

CMail * mailfrom_body();
CMail * mailfrom_file();

void sanitize(char *field);

void
sanitize(char *field) {
	if(field == NULL)
		return;

	for(; *field; field++) {
		switch(*field) {
		case '\n':
		case '\r':
			*field = ' ';
		}
	}
}

int
sendmessage() {
	FILE *fd;
	char *p;

	if(!(fd=openmta())) return -1;

	/*
	 * Sanitize fields that will be processed by sendmail.
	 */
	sanitize(subj);
	sanitize(to);
	sanitize(fullname);
	sanitize(cc);
	sanitize(bcc);

	CMail *mbody = mailfrom_body();
	CMail *mfile = mailfrom_file();

	if(!mbody) {
		mbody=mfile;
		mfile = NULL;
		if(!mbody)
			return -1;
	};

	CMail *mail=NULL;

	if(mfile) {

		mail = new CMail();
		mail->AddPart(mbody);
		mail->AddPart(mfile);
		p = (char *)sf_malloc(128);
		sprintf(p, "Content-Type: multipart/mixed; \n\tboundary=\"%s\"",
			mail->ConstructMIMEBoundary());
		mail->AddHeadLine(p);
		free(p);

	} else {

		mail = mbody;

	}

#ifdef PGP

	char *dsp;
	slist *sl = param2("pgpfunc", 0);

	int es=0;

	if(sl) {
		if(scfind(sl, "encrypt") != -1) es |= 1;
		if(scfind(sl, "sign") != -1) es |= 2;
		if(es == 2)
			es |= 8;
	}

	if(es) {

		if(!to) to = "";
		if(pgpinit(to)) {
			printf("%s", cf("pgp-init-failed"));
			if(pgpErrString)
				printf("<BR>[%s]", pgpErrString);
			return -1;
		};

		p = mail->Save();

		if(es == 10)
		p = replace(p, "\n", "\r\n");

		dsp=pgpencode(p, strlen(p), es);

		pgpcleanup();

		if(!dsp) {
			printf("%s", cf("pgp-encode-failed"));
			if(pgpErrString)
				printf("<BR>[%s]", pgpErrString);
			return -1;
		}

		// Signing
		if( es == 10 ) {

			CMail *wrapmail = new CMail();
			wrapmail->AddPart(mail);

			mail = new CMail();	// Signature.

			mail->AddHeadLine("Content-Type: application/pgp-signature");
			mail->AddHeadLine("Content-Disposition: inline");

			mail->SetBody(dsp);

			wrapmail->AddPart(mail);

			char *p = (char *)sf_malloc(128);
			sprintf(p, "Content-Type: multipart/signed; micalg=pgp-md5; \n\tprotocol=\"application/pgp-signature\"; \n\tboundary=\"%s\"",
				wrapmail->ConstructMIMEBoundary());
			wrapmail->AddHeadLine(p);
			free(p);

			mail = wrapmail;

		} else {

			CMail *wrapmail = new CMail();

			CMail *version = new CMail();
			version->AddHeadLine("Content-Type: application/pgp-encrypted");
			version->SetBody("Version: 1\n");

			wrapmail->AddPart(version);

			mail = new CMail();
			mail->AddHeadLine("Content-Type: application/octet-stream");
			mail->SetBody(dsp);

			wrapmail->AddPart(mail);


			char *p = (char *)sf_malloc(128);
			sprintf(p, "Content-Type: multipart/encrypted; \n\tprotocol=\"application/pgp-encrypted\"; \n\tboundary=\"%s\"",
				wrapmail->ConstructMIMEBoundary());
			wrapmail->AddHeadLine(p);
			free(p);

			mail = wrapmail;
		}
	}

#endif

	// Actually send mail.

	fprintf(fd, "MIME-Version: 1.0\n");
	fprintf(fd, "X-Mailer: Exaview 4.0 (Web MUA, http://www.spelio.net.ru/soft)\n");

	if(fullname &&	*fullname) fprintf(fd, "From: %s\n", fullname);
	if(to &&	*to)	fprintf(fd, "To: %s\n", to);
	if(cc &&	*cc)	fprintf(fd, "Cc: %s\n", cc);
	if(bcc &&	*bcc)	fprintf(fd, "Bcc: %s\n", bcc);
	if(subj &&	*subj)	fprintf(fd, "Subject: %s\n", subj);


	mail->Save(fd);

	fclose(fd);

	return 0;
}


char *
getmimetype(char *fname) {
	slist *sl;
	FILE *fd;

	if((sl=param2("file", 2)) && sl->count)
		return sl->list[0];

	char *ext=strrchr(fname, '.');
	if(!ext)
		goto notfound;

	fname=cfgget("mime.types");
	if(!fname)
		goto notfound;

	fd = fopen(fname, "r");
	if(!fd)
		goto notfound;

	char buf[512];
	char e[256];
	static char t[256];
	while(fgets(buf, sizeof(buf), fd)) {
		sscanf(buf, "%s %s", e, t);
		if(strcmp(ext, e)) continue;
		fclose(fd);
		return t;
	}
	fclose(fd);

notfound:

	return "application/octet-stream";
}


CMail *
mailfrom_body() {
	CMail *mail=NULL;

	mail = new CMail();

	if((cfgget("charset"))) {
		char *p=cfgget("charset");
		p=(char *)sf_malloc(64 + strlen(p));
		sprintf(p, "Content-Type: text/plain; charset=%s", cfgget("charset"));
		mail->AddHeadLine(p);
		free(p);
	} else {
		mail->AddHeadLine("Content-Type: text/plain");
	}

	mail->AddHeadLine("Content-Disposition: inline");
	mail->AddHeadLine("Content-Transfer-Encoding: 8bit");

	replace(body, "\r\n", "\n");
	mail->SetBody(replace_detach(), 1);

	return mail;
}

CMail *
mailfrom_file() {
	void *file;
	char *fname;
	size_t size;
	char *p;
	char *mt;
	slist *sl;
	char *s;


	if(!(fname=param("file")) || !*fname)
		return NULL;

	sl = param2("file", 1);
	if(!sl || !sl->count)
		return 0;

	file = sl->list[0];
	size = sl->lens[0];

	CMail *mail = new CMail();

	mt=getmimetype(fname);
	p=(char *)sf_malloc(64 + strlen(fname) + strlen(mt));
	sprintf(p, "Content-Type: %s; name=\"%s\"", mt, fname);

	mail->AddHeadLine(p);

	s = param("disposition");
	if(s == NULL || strcasecmp(s, "attachment"))
		s = "inline";
	else
		s = "attachment";

	// Previously allocated memory for p will be enough.
	sprintf(p, "Content-Disposition: %s; filename=\"%s\"", s, fname);
	mail->AddHeadLine(p);
	free(p);

	s = param("description");
	if(s && *s) {
		mail->AddHeadLine("Content-Description:", s);
	}


	int n=0; // 0: binary, 1: 7 bit text, 2: 8 bit text.

	if(strncasecmp(mt, "text/", 5)) {
		int siz=size;
		p=(char *)file;
		register int c;

		while(siz--) {
			c=*(unsigned char *)p++;
			if((c < 32 && c != '\t' && c != '\n' && c != '\r') || c == 127)
				{ n=0; break; };
			if(c > 127) n=2;
		}
	}

	switch(n) {
		case 2:
			mail->AddHeadLine("Content-Transfer-Encoding: 8bit");
			break;
		case 1:
			mail->AddHeadLine("Content-Transfer-Encoding: 7bit");
			break;
		default:
			mail->AddHeadLine("Content-Transfer-Encoding: base64");
			file = base64_encode(file, size);
			base64_detach(NULL);
	}

	mail->SetBody((char *)file);

	return mail;
}


